@@ -1,29 +1,38 @@
+import copy
 import functools
 import gym
 import logging
 import numpy as np
+import os
 import time
 import threading
-from typing import Callable, Dict, List, Optional, Tuple, Type, Union
+import tree  # pip install dm_tree
+from typing import Callable, Dict, List, Optional, Set, Tuple, Type, Union, \
+    TYPE_CHECKING
 
+import ray
 from ray.rllib.models.modelv2 import ModelV2
 from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
 from ray.rllib.models.torch.torch_action_dist import TorchDistributionWrapper
 from ray.rllib.policy.policy import Policy, LEARNER_STATS_KEY
 from ray.rllib.policy.sample_batch import SampleBatch
 from ray.rllib.policy.rnn_sequencing import pad_batch_to_sequences_of_same_size
-from ray.rllib.utils import force_list
+from ray.rllib.utils import force_list, NullContextManager
 from ray.rllib.utils.annotations import override, DeveloperAPI
+from ray.rllib.utils.deprecation import deprecation_warning
 from ray.rllib.utils.framework import try_import_torch
-from ray.rllib.utils.schedules import ConstantSchedule, PiecewiseSchedule
+from ray.rllib.utils.schedules import PiecewiseSchedule
+from ray.rllib.utils.spaces.space_utils import normalize_action
 from ray.rllib.utils.threading import with_lock
 from ray.rllib.utils.torch_ops import convert_to_non_torch_type, \
     convert_to_torch_tensor
-from ray.rllib.utils.tracking_dict import UsageTrackingDict
-from ray.rllib.utils.typing import ModelGradients, ModelWeights, \
-    TensorType, TrainerConfigDict
+from ray.rllib.utils.typing import ModelGradients, ModelWeights, TensorType, \
+    TrainerConfigDict
 
-torch, _ = try_import_torch()
+if TYPE_CHECKING:
+    from ray.rllib.evaluation import MultiAgentEpisode  # noqa
+
+torch, nn = try_import_torch()
 
 logger = logging.getLogger(__name__)
 
@@ -86,7 +95,7 @@ class TorchPolicy(Policy):
                 sampled action and its log-likelihood given Policy, ModelV2,
                 input_dict, explore, timestep, and is_training.
             action_distribution_fn (Optional[Callable[[Policy, ModelV2,
-                Dict[str, TensorType], TensorType, TensorType],
+                ModelInputDict, TensorType, TensorType],
                 Tuple[TensorType, type, List[TensorType]]]]): A callable
                 returning distribution inputs (parameters), a dist-class to
                 generate an action distribution object from, and
@@ -96,7 +105,7 @@ class TorchPolicy(Policy):
                 forward pass through some model.
                 If None, pass inputs through `self.model()` to get distribution
                 inputs.
-                The callable takes as inputs: Policy, ModelV2, input_dict,
+                The callable takes as inputs: Policy, ModelV2, ModelInputDict,
                 explore, timestep, is_training.
             max_seq_len (int): Max sequence length for LSTM training.
             get_batch_divisibility_req (Optional[Callable[[Policy], int]]]):
@@ -105,13 +114,55 @@ class TorchPolicy(Policy):
         """
         self.framework = "torch"
         super().__init__(observation_space, action_space, config)
-        if torch.cuda.is_available():
-            logger.info("TorchPolicy running on GPU.")
-            self.device = torch.device("cuda")
-        else:
-            logger.info("TorchPolicy running on CPU.")
+
+        # Log device and worker index.
+        from ray.rllib.evaluation.rollout_worker import get_global_worker
+        worker = get_global_worker()
+        worker_idx = worker.worker_index if worker else 0
+
+        # Create multi-GPU model towers, if necessary.
+        # - The central main model will be stored under self.model, residing on
+        #   self.device.
+        # - Each GPU will have a copy of that model under
+        #   self.model_gpu_towers, matching the devices in self.devices.
+        # - Parallelization is done by splitting the train batch and passing
+        #   it through the model copies in parallel, then averaging over the
+        #   resulting gradients, applying these averages on the main model and
+        #   updating all towers' weights from the main model.
+        # - In case of just one device (1 (fake) GPU or 1 CPU), no
+        #   parallelization will be done.
+        if config["_fake_gpus"] or config["num_gpus"] == 0 or \
+                not torch.cuda.is_available():
+            logger.info("TorchPolicy (worker={}) running on {}.".format(
+                worker_idx if worker_idx > 0 else "local",
+                "{} fake-GPUs".format(config["num_gpus"])
+                if config["_fake_gpus"] else "CPU"))
             self.device = torch.device("cpu")
-        self.model = model.to(self.device)
+            self.devices = [
+                self.device for _ in range(config["num_gpus"] or 1)
+            ]
+            self.model_gpu_towers = [
+                model if i == 0 else copy.deepcopy(model)
+                for i in range(config["num_gpus"] or 1)
+            ]
+            self.model = model
+        else:
+            logger.info("TorchPolicy (worker={}) running on {} GPU(s).".format(
+                worker_idx if worker_idx > 0 else "local", config["num_gpus"]))
+            gpu_ids = ray.get_gpu_ids()
+            self.devices = [
+                torch.device("cuda:{}".format(i))
+                for i, id_ in enumerate(gpu_ids) if i < config["num_gpus"]
+            ]
+            self.device = self.devices[0]
+            ids = [
+                id_ for i, id_ in enumerate(gpu_ids) if i < config["num_gpus"]
+            ]
+            self.model_gpu_towers = []
+            for i, _ in enumerate(ids):
+                model_copy = copy.deepcopy(model)
+                self.model_gpu_towers.append(model_copy.to(self.devices[i]))
+            self.model = self.model_gpu_towers[0]
 
         # Lock used for locking some methods on the object-level.
         # This prevents possible race conditions when calling the model
@@ -131,6 +182,17 @@ class TorchPolicy(Policy):
         self.unwrapped_model = model  # used to support DistributedDataParallel
         self._loss = loss
         self._optimizers = force_list(self.optimizer())
+        # Store, which params (by index within the model's list of
+        # parameters) should be updated per optimizer.
+        # Maps optimizer idx to set or param indices.
+        self.multi_gpu_param_groups: List[Set[int]] = []
+        main_params = {p: i for i, p in enumerate(self.model.parameters())}
+        for o in self._optimizers:
+            param_indices = []
+            for pg_idx, pg in enumerate(o.param_groups):
+                for p in pg["params"]:
+                    param_indices.append(main_params[p])
+            self.multi_gpu_param_groups.append(set(param_indices))
 
         self.dist_class = action_distribution_class
         self.action_sampler_fn = action_sampler_fn
@@ -159,15 +221,12 @@ class TorchPolicy(Policy):
             **kwargs) -> \
             Tuple[TensorType, List[TensorType], Dict[str, TensorType]]:
 
-        explore = explore if explore is not None else self.config["explore"]
-        timestep = timestep if timestep is not None else self.global_timestep
-
         with torch.no_grad():
             seq_lens = torch.ones(len(obs_batch), dtype=torch.int32)
-            input_dict = self._lazy_tensor_dict({
-                SampleBatch.CUR_OBS: np.asarray(obs_batch),
-                "is_training": False,
-            })
+            input_dict = self._lazy_tensor_dict(
+                SampleBatch({
+                    SampleBatch.CUR_OBS: np.asarray(obs_batch),
+                }))
             if prev_action_batch is not None:
                 input_dict[SampleBatch.PREV_ACTIONS] = \
                     np.asarray(prev_action_batch)
@@ -190,9 +249,6 @@ class TorchPolicy(Policy):
             **kwargs) -> \
             Tuple[TensorType, List[TensorType], Dict[str, TensorType]]:
 
-        explore = explore if explore is not None else self.config["explore"]
-        timestep = timestep if timestep is not None else self.global_timestep
-
         with torch.no_grad():
             # Pass lazy (torch) tensor dict to Model as `input_dict`.
             input_dict = self._lazy_tensor_dict(input_dict)
@@ -216,6 +272,8 @@ class TorchPolicy(Policy):
             Tuple:
                 - actions, state_out, extra_fetches, logp.
         """
+        explore = explore if explore is not None else self.config["explore"]
+        timestep = timestep if timestep is not None else self.global_timestep
         self._is_recurrent = state_batches is not None and state_batches != []
 
         # Switch to eval mode.
@@ -236,14 +294,34 @@ class TorchPolicy(Policy):
             self.exploration.before_compute_actions(
                 explore=explore, timestep=timestep)
             if self.action_distribution_fn:
-                dist_inputs, dist_class, state_out = \
-                    self.action_distribution_fn(
-                        self,
-                        self.model,
-                        input_dict[SampleBatch.CUR_OBS],
-                        explore=explore,
-                        timestep=timestep,
-                        is_training=False)
+                # Try new action_distribution_fn signature, supporting
+                # state_batches and seq_lens.
+                try:
+                    dist_inputs, dist_class, state_out = \
+                        self.action_distribution_fn(
+                            self,
+                            self.model,
+                            input_dict=input_dict,
+                            state_batches=state_batches,
+                            seq_lens=seq_lens,
+                            explore=explore,
+                            timestep=timestep,
+                            is_training=False)
+                # Trying the old way (to stay backward compatible).
+                # TODO: Remove in future.
+                except TypeError as e:
+                    if "positional argument" in e.args[0] or \
+                            "unexpected keyword argument" in e.args[0]:
+                        dist_inputs, dist_class, state_out = \
+                            self.action_distribution_fn(
+                                self,
+                                self.model,
+                                input_dict[SampleBatch.CUR_OBS],
+                                explore=explore,
+                                timestep=timestep,
+                                is_training=False)
+                    else:
+                        raise e
             else:
                 dist_class = self.dist_class
                 dist_inputs, state_out = self.model(input_dict, state_batches,
@@ -296,8 +374,10 @@ class TorchPolicy(Policy):
             state_batches: Optional[List[TensorType]] = None,
             prev_action_batch: Optional[Union[List[TensorType],
                                               TensorType]] = None,
-            prev_reward_batch: Optional[Union[List[
-                TensorType], TensorType]] = None) -> TensorType:
+            prev_reward_batch: Optional[Union[List[TensorType],
+                                              TensorType]] = None,
+            actions_normalized: bool = True,
+    ) -> TensorType:
 
         if self.action_sampler_fn and self.action_distribution_fn is None:
             raise ValueError("Cannot compute log-prob/likelihood w/o an "
@@ -324,12 +404,34 @@ class TorchPolicy(Policy):
 
             # Action dist class and inputs are generated via custom function.
             if self.action_distribution_fn:
-                dist_inputs, dist_class, _ = self.action_distribution_fn(
-                    policy=self,
-                    model=self.model,
-                    obs_batch=input_dict[SampleBatch.CUR_OBS],
-                    explore=False,
-                    is_training=False)
+
+                # Try new action_distribution_fn signature, supporting
+                # state_batches and seq_lens.
+                try:
+                    dist_inputs, dist_class, state_out = \
+                        self.action_distribution_fn(
+                            self,
+                            self.model,
+                            input_dict=input_dict,
+                            state_batches=state_batches,
+                            seq_lens=seq_lens,
+                            explore=False,
+                            is_training=False)
+                # Trying the old way (to stay backward compatible).
+                # TODO: Remove in future.
+                except TypeError as e:
+                    if "positional argument" in e.args[0] or \
+                            "unexpected keyword argument" in e.args[0]:
+                        dist_inputs, dist_class, _ = \
+                            self.action_distribution_fn(
+                                policy=self,
+                                model=self.model,
+                                obs_batch=input_dict[SampleBatch.CUR_OBS],
+                                explore=False,
+                                is_training=False)
+                    else:
+                        raise e
+
             # Default action-dist inputs calculation.
             else:
                 dist_class = self.dist_class
@@ -337,7 +439,13 @@ class TorchPolicy(Policy):
                                             seq_lens)
 
             action_dist = dist_class(dist_inputs, self.model)
-            log_likelihoods = action_dist.logp(input_dict[SampleBatch.ACTIONS])
+
+            # Normalize actions if necessary.
+            actions = input_dict[SampleBatch.ACTIONS]
+            if not actions_normalized and self.config["normalize_actions"]:
+                actions = normalize_action(actions, self.action_space_struct)
+
+            log_likelihoods = action_dist.logp(actions)
 
             return log_likelihoods
 
@@ -351,19 +459,20 @@ class TorchPolicy(Policy):
         if self.model:
             self.model.train()
         # Callback handling.
+        learn_stats = {}
         self.callbacks.on_learn_on_batch(
-            policy=self, train_batch=postprocessed_batch)
+            policy=self, train_batch=postprocessed_batch, result=learn_stats)
 
         # Compute gradients (will calculate all losses and `backward()`
         # them to get the grads).
         grads, fetches = self.compute_gradients(postprocessed_batch)
 
         # Step the optimizers.
-        for i, opt in enumerate(self._optimizers):
-            opt.step()
+        self.apply_gradients(_directStepOptimizerSingleton)
 
         if self.model:
             fetches["model"] = self.model.metrics()
+        fetches.update({"custom_metrics": learn_stats})
 
         return fetches
 
@@ -373,93 +482,92 @@ class TorchPolicy(Policy):
     def compute_gradients(self,
                           postprocessed_batch: SampleBatch) -> ModelGradients:
 
-        pad_batch_to_sequences_of_same_size(
-            postprocessed_batch,
-            max_seq_len=self.max_seq_len,
-            shuffle=False,
-            batch_divisibility_req=self.batch_divisibility_req,
-            view_requirements=self.view_requirements,
-        )
+        # For multi-GPU, split the batch into n slices (n=#GPUs).
+        if len(self.devices) == 1:
+            batches = [postprocessed_batch]
+        else:
+            from ray.rllib.utils.sgd import minibatches
+            batches = list(
+                minibatches(
+                    postprocessed_batch,
+                    len(postprocessed_batch) // len(self.devices),
+                    shuffle=False))
 
-        # Mark the batch as "is_training" so the Model can use this
-        # information.
-        postprocessed_batch["is_training"] = True
-        train_batch = self._lazy_tensor_dict(postprocessed_batch)
+        if not isinstance(postprocessed_batch, SampleBatch) or \
+                not postprocessed_batch.zero_padded:
+            for b in batches:
+                pad_batch_to_sequences_of_same_size(
+                    b,
+                    max_seq_len=self.max_seq_len,
+                    shuffle=False,
+                    batch_divisibility_req=self.batch_divisibility_req,
+                    view_requirements=self.view_requirements,
+                )
 
-        # Calculate the actual policy loss.
-        loss_out = force_list(
-            self._loss(self, self.model, self.dist_class, train_batch))
+        for b, d in zip(batches, self.devices):
+            b.is_training = True
+            self._lazy_tensor_dict(b, device=d)
 
-        # Call Model's custom-loss with Policy loss outputs and train_batch.
-        if self.model:
-            loss_out = self.model.custom_loss(loss_out, train_batch)
+        # Multi-GPU case: Slice inputs into n (roughly) equal batches.
+        if len(self.devices) > 1:
+            # Copy weights of main model to all towers.
+            state_dict = self.model.state_dict()
+            for tower in self.model_gpu_towers:
+                tower.load_state_dict(state_dict)
 
-        # Give Exploration component that chance to modify the loss (or add
-        # its own terms).
-        if hasattr(self, "exploration"):
-            loss_out = self.exploration.get_exploration_loss(
-                loss_out, train_batch)
+        # Do the (maybe parallelized) gradient calculation step.
+        tower_outputs = self._multi_gpu_parallel_grad_calc(batches)
 
-        assert len(loss_out) == len(self._optimizers)
-
-        # assert not any(torch.isnan(l) for l in loss_out)
-        fetches = self.extra_compute_grad_fetches()
-
-        # Loop through all optimizers.
-        grad_info = {"allreduce_latency": 0.0}
-
-        all_grads = []
-        for i, opt in enumerate(self._optimizers):
-            # Erase gradients in all vars of this optimizer.
-            opt.zero_grad()
-            # Recompute gradients of loss over all variables.
-            loss_out[i].backward(retain_graph=(i < len(self._optimizers) - 1))
-            grad_info.update(self.extra_grad_process(opt, loss_out[i]))
-
-            grads = []
-            # Note that return values are just references;
-            # Calling zero_grad would modify the values.
-            for param_group in opt.param_groups:
-                for p in param_group["params"]:
-                    if p.grad is not None:
-                        grads.append(p.grad)
-                        all_grads.append(p.grad.data.cpu().numpy())
-                    else:
-                        all_grads.append(None)
-
-            if self.distributed_world_size:
-                start = time.time()
-                if torch.cuda.is_available():
-                    # Sadly, allreduce_coalesced does not work with CUDA yet.
-                    for g in grads:
-                        torch.distributed.all_reduce(
-                            g, op=torch.distributed.ReduceOp.SUM)
+        # Multi device (GPU) case.
+        if len(self.devices) > 1:
+            # Mean-reduce over GPU-towers.
+            all_grads = []
+            for i in range(len(tower_outputs[0][0])):
+                if tower_outputs[0][0][i] is not None:
+                    all_grads.append(
+                        torch.mean(
+                            torch.stack([
+                                t[0][i].to(self.device) for t in tower_outputs
+                            ]),
+                            dim=0))
                 else:
-                    torch.distributed.all_reduce_coalesced(
-                        grads, op=torch.distributed.ReduceOp.SUM)
-
-                for param_group in opt.param_groups:
-                    for p in param_group["params"]:
-                        if p.grad is not None:
-                            p.grad /= self.distributed_world_size
-
-                grad_info["allreduce_latency"] += time.time() - start
+                    all_grads.append(None)
+            # Set main model's grads to mean-reduced values.
+            for i, p in enumerate(self.model.parameters()):
+                p.grad = all_grads[i]
+            # Reduce stats over towers as well.
+            from ray.rllib.execution.train_ops import all_tower_reduce
+            grad_info = tree.map_structure_with_path(
+                lambda p, *t: all_tower_reduce(p, *t),
+                *[t[1] for t in tower_outputs])
+        # Single device case.
+        else:
+            all_grads, grad_info = tower_outputs[0]
 
         grad_info["allreduce_latency"] /= len(self._optimizers)
-        grad_info.update(self.extra_grad_info(train_batch))
+        grad_info.update(self.extra_grad_info(postprocessed_batch))
+
+        fetches = self.extra_compute_grad_fetches()
 
         return all_grads, dict(fetches, **{LEARNER_STATS_KEY: grad_info})
 
     @override(Policy)
     @DeveloperAPI
     def apply_gradients(self, gradients: ModelGradients) -> None:
-        # TODO(sven): Not supported for multiple optimizers yet.
-        assert len(self._optimizers) == 1
-        for g, p in zip(gradients, self.model.parameters()):
-            if g is not None:
-                p.grad = torch.from_numpy(g).to(self.device)
+        if gradients == _directStepOptimizerSingleton:
+            for i, opt in enumerate(self._optimizers):
+                opt.step()
+        else:
+            # TODO(sven): Not supported for multiple optimizers yet.
+            assert len(self._optimizers) == 1
+            for g, p in zip(gradients, self.model.parameters()):
+                if g is not None:
+                    if torch.is_tensor(g):
+                        p.grad = g.to(self.device)
+                    else:
+                        p.grad = torch.from_numpy(g).to(self.device)
 
-        self._optimizers[0].step()
+            self._optimizers[0].step()
 
     @override(Policy)
     @DeveloperAPI
@@ -500,20 +608,25 @@ class TorchPolicy(Policy):
         for i, o in enumerate(self._optimizers):
             optim_state_dict = convert_to_non_torch_type(o.state_dict())
             state["_optimizer_variables"].append(optim_state_dict)
+        # Add exploration state.
+        state["_exploration_state"] = \
+            self.exploration.get_state()
         return state
 
     @override(Policy)
     @DeveloperAPI
-    def set_state(self, state: object) -> None:
-        state = state.copy()  # shallow copy
+    def set_state(self, state: dict) -> None:
         # Set optimizer vars first.
-        optimizer_vars = state.pop("_optimizer_variables", None)
+        optimizer_vars = state.get("_optimizer_variables", None)
         if optimizer_vars:
             assert len(optimizer_vars) == len(self._optimizers)
             for o, s in zip(self._optimizers, optimizer_vars):
                 optim_state_dict = convert_to_torch_tensor(
                     s, device=self.device)
                 o.load_state_dict(optim_state_dict)
+        # Set exploration's state.
+        if hasattr(self, "exploration") and "_exploration_state" in state:
+            self.exploration.set_state(state=state["_exploration_state"])
         # Then the Policy's (NN) weights.
         super().set_state(state)
 
@@ -600,15 +713,41 @@ class TorchPolicy(Policy):
     @override(Policy)
     @DeveloperAPI
     def export_model(self, export_dir: str) -> None:
-        """TODO(sven): implement for torch.
-        """
-        raise NotImplementedError
+        """Exports the Policy's Model to local directory for serving.
 
-    @override(Policy)
-    @DeveloperAPI
-    def export_checkpoint(self, export_dir: str) -> None:
-        """TODO(sven): implement for torch.
+        Creates a TorchScript model and saves it.
+
+        Args:
+            export_dir (str): Local writable directory or filename.
         """
+        self._lazy_tensor_dict(self._dummy_batch)
+        # Provide dummy state inputs if not an RNN (torch cannot jit with
+        # returned empty internal states list).
+        if "state_in_0" not in self._dummy_batch:
+            self._dummy_batch["state_in_0"] = \
+                self._dummy_batch["seq_lens"] = np.array([1.0])
+        seq_lens = self._dummy_batch["seq_lens"]
+
+        state_ins = []
+        i = 0
+        while "state_in_{}".format(i) in self._dummy_batch:
+            state_ins.append(self._dummy_batch["state_in_{}".format(i)])
+            i += 1
+        dummy_inputs = {
+            k: self._dummy_batch[k]
+            for k in self._dummy_batch.keys() if k != "is_training"
+        }
+        traced = torch.jit.trace(self.model,
+                                 (dummy_inputs, state_ins, seq_lens))
+        if not os.path.exists(export_dir):
+            os.makedirs(export_dir)
+        file_name = os.path.join(export_dir, "model.pt")
+        traced.save(file_name)
+
+    # TODO: (sven) Deprecate this in favor of `save()`.
+    @override(Policy)
+    def export_checkpoint(self, export_dir: str) -> None:
+        deprecation_warning("export_checkpoint", "save")
         raise NotImplementedError
 
     @override(Policy)
@@ -617,17 +756,138 @@ class TorchPolicy(Policy):
         """Imports weights into torch model."""
         return self.model.import_from_h5(import_file)
 
-    def _lazy_tensor_dict(self, postprocessed_batch):
-        train_batch = UsageTrackingDict(postprocessed_batch)
-        train_batch.set_get_interceptor(
-            functools.partial(convert_to_torch_tensor, device=self.device))
-        return train_batch
+    def _lazy_tensor_dict(self, postprocessed_batch: SampleBatch, device=None):
+        # TODO: (sven): Keep for a while to ensure backward compatibility.
+        if not isinstance(postprocessed_batch, SampleBatch):
+            postprocessed_batch = SampleBatch(postprocessed_batch)
+        postprocessed_batch.set_get_interceptor(
+            functools.partial(
+                convert_to_torch_tensor, device=device or self.device))
+        return postprocessed_batch
 
-    def _lazy_numpy_dict(self, postprocessed_batch):
-        train_batch = UsageTrackingDict(postprocessed_batch)
-        train_batch.set_get_interceptor(
-            functools.partial(convert_to_non_torch_type))
-        return train_batch
+    def _multi_gpu_parallel_grad_calc(self, sample_batches):
+        """Performs a parallelized loss and gradient calculation over the batch.
+
+        Splits up the given train batch into n shards (n=number of this
+        Policy's devices) and passes each data shard (in parallel) through
+        the loss function using the individual devices' models
+        (self.model_gpu_towers). Then returns each tower's outputs.
+
+        Args:
+            sample_batches (List[SampleBatch]): A list of SampleBatch shards to
+                calculate loss and gradients for.
+
+        Returns:
+            List[Tuple[List[TensorType], StatsDict]]: A list (one item per
+                device) of 2-tuples with 1) gradient list and 2) stats dict.
+        """
+        assert len(self.model_gpu_towers) == len(sample_batches)
+        lock = threading.Lock()
+        results = {}
+        grad_enabled = torch.is_grad_enabled()
+
+        def _worker(shard_idx, model, sample_batch, device):
+            torch.set_grad_enabled(grad_enabled)
+            try:
+                with NullContextManager(
+                ) if device.type == "cpu" else torch.cuda.device(device):
+                    loss_out = force_list(
+                        self._loss(self, model, self.dist_class, sample_batch))
+
+                    # Call Model's custom-loss with Policy loss outputs and
+                    # train_batch.
+                    loss_out = model.custom_loss(loss_out, sample_batch)
+
+                    assert len(loss_out) == len(self._optimizers)
+
+                    # Loop through all optimizers.
+                    grad_info = {"allreduce_latency": 0.0}
+
+                    parameters = list(model.parameters())
+                    all_grads = [None for _ in range(len(parameters))]
+                    for opt_idx, opt in enumerate(self._optimizers):
+                        # Erase gradients in all vars of the tower that this
+                        # optimizer would affect.
+                        param_indices = self.multi_gpu_param_groups[opt_idx]
+                        for param_idx, param in enumerate(parameters):
+                            if param_idx in param_indices and \
+                                    param.grad is not None:
+                                param.grad.data.zero_()
+                        # Recompute gradients of loss over all variables.
+                        loss_out[opt_idx].backward(retain_graph=True)
+                        grad_info.update(
+                            self.extra_grad_process(opt, loss_out[opt_idx]))
+
+                        grads = []
+                        # Note that return values are just references;
+                        # Calling zero_grad would modify the values.
+                        for param_idx, param in enumerate(parameters):
+                            if param_idx in param_indices:
+                                if param.grad is not None:
+                                    grads.append(param.grad)
+                                all_grads[param_idx] = param.grad
+
+                        if self.distributed_world_size:
+                            start = time.time()
+                            if torch.cuda.is_available():
+                                # Sadly, allreduce_coalesced does not work with
+                                # CUDA yet.
+                                for g in grads:
+                                    torch.distributed.all_reduce(
+                                        g, op=torch.distributed.ReduceOp.SUM)
+                            else:
+                                torch.distributed.all_reduce_coalesced(
+                                    grads, op=torch.distributed.ReduceOp.SUM)
+
+                            for param_group in opt.param_groups:
+                                for p in param_group["params"]:
+                                    if p.grad is not None:
+                                        p.grad /= self.distributed_world_size
+
+                            grad_info[
+                                "allreduce_latency"] += time.time() - start
+
+                with lock:
+                    results[shard_idx] = (all_grads, grad_info)
+            except Exception as e:
+                with lock:
+                    results[shard_idx] = ValueError(
+                        e.args[0] + "\n" +
+                        "In tower {} on device {}".format(shard_idx, device))
+
+        # Single device (GPU) or fake-GPU case (serialize for better
+        # debugging).
+        if len(self.devices) == 1 or self.config["_fake_gpus"]:
+            for shard_idx, (model, sample_batch, device) in enumerate(
+                    zip(self.model_gpu_towers, sample_batches, self.devices)):
+                _worker(shard_idx, model, sample_batch, device)
+                # Raise errors right away for better debugging.
+                last_result = results[len(results) - 1]
+                if isinstance(last_result, ValueError):
+                    raise last_result
+        # Multi device (GPU) case: Parallelize via threads.
+        else:
+            threads = [
+                threading.Thread(
+                    target=_worker,
+                    args=(shard_idx, model, sample_batch, device))
+                for shard_idx, (model, sample_batch, device) in enumerate(
+                    zip(self.model_gpu_towers, sample_batches, self.devices))
+            ]
+
+            for thread in threads:
+                thread.start()
+            for thread in threads:
+                thread.join()
+
+        # Gather all threads' outputs and return.
+        outputs = []
+        for shard_idx in range(len(sample_batches)):
+            output = results[shard_idx]
+            if isinstance(output, Exception):
+                raise output
+            outputs.append(results[shard_idx])
+        return outputs
 
 
 # TODO: (sven) Unify hyperparam annealing procedures across RLlib (tf/torch)
@@ -638,20 +898,22 @@ class LearningRateSchedule:
 
     @DeveloperAPI
     def __init__(self, lr, lr_schedule):
-        self.cur_lr = lr
+        self._lr_schedule = None
         if lr_schedule is None:
-            self.lr_schedule = ConstantSchedule(lr, framework=None)
+            self.cur_lr = lr
         else:
-            self.lr_schedule = PiecewiseSchedule(
+            self._lr_schedule = PiecewiseSchedule(
                 lr_schedule, outside_value=lr_schedule[-1][-1], framework=None)
+            self.cur_lr = self._lr_schedule.value(0)
 
     @override(Policy)
     def on_global_var_update(self, global_vars):
         super().on_global_var_update(global_vars)
-        self.cur_lr = self.lr_schedule.value(global_vars["timestep"])
-        for opt in self._optimizers:
-            for p in opt.param_groups:
-                p["lr"] = self.cur_lr
+        if self._lr_schedule:
+            self.cur_lr = self._lr_schedule.value(global_vars["timestep"])
+            for opt in self._optimizers:
+                for p in opt.param_groups:
+                    p["lr"] = self.cur_lr
 
 
 @DeveloperAPI
@@ -660,27 +922,49 @@ class EntropyCoeffSchedule:
 
     @DeveloperAPI
     def __init__(self, entropy_coeff, entropy_coeff_schedule):
-        self.entropy_coeff = entropy_coeff
-
+        self._entropy_coeff_schedule = None
         if entropy_coeff_schedule is None:
-            self.entropy_coeff_schedule = ConstantSchedule(
-                entropy_coeff, framework=None)
+            self.entropy_coeff = entropy_coeff
         else:
             # Allows for custom schedule similar to lr_schedule format
             if isinstance(entropy_coeff_schedule, list):
-                self.entropy_coeff_schedule = PiecewiseSchedule(
+                self._entropy_coeff_schedule = PiecewiseSchedule(
                     entropy_coeff_schedule,
                     outside_value=entropy_coeff_schedule[-1][-1],
                     framework=None)
             else:
                 # Implements previous version but enforces outside_value
-                self.entropy_coeff_schedule = PiecewiseSchedule(
+                self._entropy_coeff_schedule = PiecewiseSchedule(
                     [[0, entropy_coeff], [entropy_coeff_schedule, 0.0]],
                     outside_value=0.0,
                     framework=None)
+            self.entropy_coeff = self._entropy_coeff_schedule.value(0)
 
     @override(Policy)
     def on_global_var_update(self, global_vars):
         super(EntropyCoeffSchedule, self).on_global_var_update(global_vars)
-        self.entropy_coeff = self.entropy_coeff_schedule.value(
-            global_vars["timestep"])
+        if self._entropy_coeff_schedule is not None:
+            self.entropy_coeff = self._entropy_coeff_schedule.value(
+                global_vars["timestep"])
+
+
+@DeveloperAPI
+class DirectStepOptimizer:
+    """Typesafe method for indicating apply gradients can directly step the
+       optimizers with in-place gradients.
+    """
+    _instance = None
+
+    def __new__(cls):
+        if DirectStepOptimizer._instance is None:
+            DirectStepOptimizer._instance = super().__new__(cls)
+        return DirectStepOptimizer._instance
+
+    def __eq__(self, other):
+        return type(self) == type(other)
+
+    def __repr__(self):
+        return "DirectStepOptimizer"
+
+
+_directStepOptimizerSingleton = DirectStepOptimizer()
